/*
 * (c) 2007 Stanislaw K Skowronek
 * Licensed under the terms of GPLv2 or later
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <linux/ppdev.h>
#include <linux/ioctl.h>

#define MAX_IR 2048
#define MAX_DEV 80

int fd;
int total_ir;
int ndev;
int bypass[MAX_DEV];
char *name[MAX_DEV];
int irlen[MAX_DEV];
uint32_t sample[MAX_DEV];
uint32_t extest[MAX_DEV];
int bslen[MAX_DEV];
char **bsname[MAX_DEV];
char *bsmode[MAX_DEV];
int *bszpin[MAX_DEV];
int *bszval[MAX_DEV];
int *bsextr[MAX_DEV];

int *bsc_in[MAX_DEV];
int *bsc_out[MAX_DEV];

int *bsc_tmp[MAX_DEV];
int *bsc_clk[MAX_DEV];
int *bsc_sig0[MAX_DEV];
int *bsc_sig1[MAX_DEV];

int init(char *name)
{
	fd=open(name, O_RDWR);
	if(fd<0)
		return 1;
	if((ioctl(fd, PPEXCL)==-1) || (ioctl(fd, PPCLAIM)==-1)) {
		close(fd);
		return 1;
	}
	return 0;
}

void done(void)
{
	ioctl(fd, PPRELEASE);
	close(fd);
}

void setdata(uint8_t val)
{
	ioctl(fd, PPWDATA, &val);
	ioctl(fd, PPWDATA, &val);
}

uint8_t getstat(void)
{
	uint8_t res;
	ioctl(fd, PPRSTATUS, &res);
	return res^0x80;
}

int findcable(void)
{
	int i;
	for(i=0;i<16;i++) {
		setdata(0x40);
		if(!(getstat()&0x80))
			return 1;
		setdata(0x00);
		if((getstat()&0x80))
			return 1;
	}
	return 0;
}

void clock(int tms, int tdi)
{
	setdata(0x10|(tms?4:0)|(tdi?1:0));
	setdata(0x12|(tms?4:0)|(tdi?1:0));
}

int gettdo(void)
{
	setdata(0x10);
	return (getstat()>>4)&1;
}

void reset(void)
{
	int i;
	for(i=0;i<5;i++)
		clock(1,0);
	clock(0,0);
}

void cap_dr(void)
{
	clock(1,0);
	clock(0,0);
}

void cap_ir(void)
{
	clock(1,0);
	clock(1,0);
	clock(0,0);
}

int regsize(int max)
{
	int i;
	for(i=0;i<max;i++)
		clock(0,1);
	clock(0,0);
	for(i=0;i<max;i++) {
		if(!gettdo()) {
			clock(1,1);
			clock(1,1);
			clock(0,1);
			return i+1;
		}
		clock(0,1);
	}
	reset();
	return -1;
}

int size_chain(void)
{
	cap_ir();
	total_ir=regsize(MAX_IR);
	if(total_ir==-1)
		return 1;
	cap_dr();
	ndev=regsize(MAX_DEV);
	if(ndev==-1)
		return 1;
	return 0;
}

void loadinfo(int i, uint32_t b)
{
	char fname[32];
	char line[256], *p, *q;
	int idx;
	FILE *f;
	if(b!=0)
	    sprintf(fname, "dev/%08X", b);
	else
	    sprintf(fname, "local/%d", i);
	f=fopen(fname,"r");
	if(!f) {
		name[i]=NULL;
		return;
	}
	bypass[i]=7;
	while(fgets(line, 256, f)) {
		p=strpbrk(line, "\n\r");
		if(p)
			*p=0;
		p=strpbrk(line, "\t ");
		if(!p) {
			fprintf(stderr, "%s: malformed line '%s'\n", fname, line);
			name[i]=NULL;
			return;
		}
		*p=0;
		p++;
		while(*p=='\t' || *p==' ')
			p++;
		if(!strcmp(line,"name")) {
			name[i]=strdup(p);
		} else if(!strcmp(line,"irsize")) {
			irlen[i]=strtol(p,NULL,0);
		} else if(!strcmp(line,"sample")) {
			sample[i]=strtol(p,NULL,2);
			bypass[i]&=~1;
		} else if(!strcmp(line,"extest")) {
			extest[i]=strtol(p,NULL,2);
			bypass[i]&=~2;
		} else if(!strcmp(line,"bssize")) {
			bslen[i]=strtol(p,NULL,0);
			bsc_in[i]=calloc(sizeof(int), bslen[i]);
			bsc_out[i]=calloc(sizeof(int), bslen[i]);
			bypass[i]&=~4;
			bsname[i]=calloc(sizeof(char *), bslen[i]);
			bsmode[i]=calloc(sizeof(char), bslen[i]);
			bszpin[i]=calloc(sizeof(int), bslen[i]);
			bszval[i]=calloc(sizeof(int), bslen[i]);
			bsextr[i]=calloc(sizeof(int), bslen[i]);
		} else if(!strncmp(line,"bsc[",4) || !strncmp(line,"ext[",4)) {
			idx=strtol(line+4,&q,0);
			if(!q || *q!=']' || idx<0 || idx>bslen[i]) {
				fprintf(stderr, "%s: bad index in '%s'\n", fname, line);
				name[i]=NULL;
				return;
			}
			q=strchr(p,' ');
			*(q++)=0;
			if(!strncmp(line,"ext[",4))
				bsextr[i][idx]=1;
			bsname[i][idx]=strdup(p);
			bsmode[i][idx]=*q;
			if(*q=='Z' || *q=='Y') {
				bszval[i][idx]=q[1]-'0';
				bszpin[i][idx]=strtol(q+3, &p, 0);
				if(!p || *p) {
					fprintf(stderr, "%s: bad tristate pin in '%s'\n", fname, line);
					name[i]=NULL;
					return;
				}
			}
		} else {
			fprintf(stderr, "%s: unknown field '%s'\n", fname, line);
			name[i]=NULL;
			return;
		}
	}
	fclose(f);
}

int idcodes(void)
{
	int i,j,fail=0,irsum=0,alt=0;
	uint32_t b;
	reset();
	cap_dr();
	for(i=0;i<ndev;i++) {
		clock(0,alt);
		alt^=1;
		b=gettdo();
		if(b)
			for(j=1;j<32;j++) {
				clock(0,alt);
				alt^=1;
				b|=gettdo()<<j;
			}
		loadinfo(i, 0);
		if(!name[i])
		    loadinfo(i, b);
		if(name[i]) {
			fprintf(stderr, "Device %d: IDCODE %08X (%s)\n", i, b, name[i]);
			if(!irlen[i]) {
				fprintf(stderr, "  Missing IR length\n");
				fail=1;
			}
			if(bypass[i])
				fprintf(stderr, "  Device is bypassed\n");
			irsum+=irlen[i];
		} else {
			fprintf(stderr, "Device %d: IDCODE %08X (MANUF: %03X, PART: %04X, VER: %01X)\n", i, b, (b>>1)&0x3FF, (b>>12)&0xFFFF, (b>>28)&0xF);
			fail=1;
		}
	}
	clock(1,1);
	clock(1,1);
	clock(0,1);
	if(irsum!=total_ir) {
		fprintf(stderr, "Total IR length %d mismatches sum of device IR lengths %d\n", total_ir, irsum);
		fail=1;
	}
	return fail;
}

uint32_t shifti(uint32_t v, int l, int t)
{
	uint32_t r=0,s=(1<<(l-1));
	while(l) {
		--l;
		r>>=1;
		r|=gettdo()?s:0;
		clock(!l?t:0,v&1);
		v>>=1;
	}
	return r;
}

void idle(void)
{
	clock(1,0);
	clock(0,0);
	clock(0,0);
	clock(0,0);
	clock(0,0);
}

void set_ir_all(uint32_t *ir)
{
	int i;
	cap_ir();
	clock(0,1);
	for(i=0;i<ndev;i++) {
		if(bypass[i] || !ir)
			shifti(-1, irlen[i], i==(ndev-1));
		else
			shifti(ir[i], irlen[i], i==(ndev-1));
	}
	idle();
}

void bsc_to_allz(void)
{
	int i,j;
	for(i=0;i<ndev;i++)
		if(!bypass[i]) {
			for(j=0;j<bslen[i];j++)
				bsc_out[i][j]=0;
			for(j=0;j<bslen[i];j++)
				if(bsmode[i][j]=='Y' || bsmode[i][j]=='Z')
					bsc_out[i][bszpin[i][j]]=bszval[i][j];
		}
}

void shift_bsc(void)
{
	int i,j;
	cap_dr();
	clock(0,0);
	for(i=0;i<ndev;i++) {
		if(bypass[i]) {
			clock(i==(ndev-1), 1);
		} else {
			for(j=0;j<bslen[i];j++) {
				bsc_in[i][j]=gettdo();
				clock(j==(bslen[i]-1)?i==(ndev-1):0, bsc_out[i][j]);
			}
		}
	}
	idle();
}

void print_bsc(int **bsc)
{
	int i,j,h;
	cap_dr();
	for(i=0;i<ndev;i++) {
		if(bypass[i]) {
			printf("%d: BYPASS\n", i);
		} else {
			printf("%d: ", i);
			h=0;
			for(j=0;j<bslen[i];j++)
				if(bsname[i][j]) {
					if(h)
						printf("   ");
					switch(bsmode[i][j]) {
					case 'Y':
					case 'Z':
						if(bsc[i][bszpin[i][j]]==bszval[i][j])
							printf("%s out: Z\n", bsname[i][j]);
						else
							printf("%s out: %d\n", bsname[i][j], bsc[i][j]);
						break;
					default:
						printf("%s: %d\n", bsname[i][j], bsc[i][j]);
					}
					h=1;
				}
		}
	}
}

int main(void)
{
	if(init("/dev/parport0")) {
		fprintf(stderr, "Port open failed\n");
		return 1;
	}
	if(findcable()) {
		fprintf(stderr, "Cable not found\n");
		done();
		return 1;
	}
	reset();
	if(size_chain()) {
		fprintf(stderr, "Chain sizing failed\n");
		reset();
		done();
		return 1;
	}
	fprintf(stderr, "Found %d devices with total IR length of %d\n", ndev, total_ir);
	if(idcodes()) {
		fprintf(stderr, "Missing device information\n");
		reset();
		done();
		return 1;
	}
	reset();
	bsc_to_allz();
	set_ir_all(sample);
	shift_bsc();
	print_bsc(bsc_in);
	reset();
	done();
	return 0;
}
