/*
 * (c) 2007 Stanislaw K Skowronek
 * Licensed under the terms of GPLv2 or later
 */

#include <stdio.h>
#include <malloc.h>
#include <GL/gl.h>

#include "font_regular.h"
#include "font_bold.h"

void *vbf_regular=NULL;
void *vbf_bold=NULL;

struct vbfheader {
    struct {
	unsigned char x,y,w,h;
    } metrics[256];
    unsigned char picheight;
};

struct vbffont {
    int buf_len;
    struct {
	float x1,y1,x2,y2;
	int w,h;
    } metrics[256];
    unsigned char *teximage;
    GLuint texid;
};

void *load_vbf(const unsigned char *data)
{
    struct vbfheader *hdr=(struct vbfheader *)data;
    struct vbffont *font=malloc(sizeof(struct vbffont));
    unsigned char *pix;
    int i,j,w,b;
    GLuint id;
    font->buf_len=sizeof(struct vbffont)-4;
    data+=1025;
    for(i=0;i<256;i++) {
	font->metrics[i].x1=((float)hdr->metrics[i].x)/256.0f;
	font->metrics[i].y1=((float)hdr->metrics[i].y)/256.0f;
	font->metrics[i].x2=((float)(hdr->metrics[i].x+hdr->metrics[i].w))/256.0f;
	font->metrics[i].y2=((float)(hdr->metrics[i].y+hdr->metrics[i].h))/256.0f;
	font->metrics[i].w=hdr->metrics[i].w;
	font->metrics[i].h=hdr->metrics[i].h;
    }
    pix=font->teximage=calloc(256,256);
    for(i=0;i<hdr->picheight;i++)
	for(j=0;j<32;j++) {
	    b=*(data++);
	    for(w=0;w<8;w++) {
    		*(pix++)=b&128?255:0;
		b<<=1;
	    }
	}
    glGenTextures(1,&id);
    glBindTexture(GL_TEXTURE_2D,id);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
    glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
    glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
    glTexImage2D(GL_TEXTURE_2D,0,GL_INTENSITY8,256,256,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,font->teximage);
    font->texid=id;
    return font;
}

void print_vbf(float x1, float y1, char *s, void *f, int c, float sc)
{
    unsigned char *t=(unsigned char *)s;
    struct vbffont *vf=(struct vbffont *)f;
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,vf->texid);
    glBegin(GL_QUADS);
    glColor3ub((c>>16)&255,(c>>8)&255,c&255);
    while(*t) {
        glTexCoord2f(vf->metrics[(int)*t].x1,vf->metrics[(int)*t].y1);
	glVertex3f(x1,y1,-1.0f);
        glTexCoord2f(vf->metrics[(int)*t].x2,vf->metrics[(int)*t].y1);
        glVertex3f(x1+sc*vf->metrics[(int)*t].w,y1,-1.0f);
        glTexCoord2f(vf->metrics[(int)*t].x2,vf->metrics[(int)*t].y2);
	glVertex3f(x1+sc*vf->metrics[(int)*t].w,y1-sc*vf->metrics[(int)*t].h,-1.0f);
        glTexCoord2f(vf->metrics[(int)*t].x1,vf->metrics[(int)*t].y2);
	glVertex3f(x1,y1-sc*vf->metrics[(int)*t].h,-1.0f);
	x1+=sc*vf->metrics[(int)*t].w;
	t++;
    }
    glEnd();
    glDisable(GL_TEXTURE_2D);
}

float vbf_width(char *s, void *f, float sc)
{
    unsigned char *t=(unsigned char *)s;
    struct vbffont *vf=(struct vbffont *)f;
    float x=0;
    if(!vf)
	return 0.0f;
    while(*t) {
	x+=vf->metrics[(int)*t].w;
	t++;
    }
    x*=sc;
    return x;
}

float vbf_height(char *s, void *f, float sc)
{
    unsigned char *t=(unsigned char *)s;
    struct vbffont *vf=(struct vbffont *)f;
    float y=1;
    if(!vf)
	return 0.0f;
    if(vf->metrics[32].h>y)
	y=vf->metrics[32].h;
    while(*t) {
	if(y<vf->metrics[(int)*t].h)
	    y=vf->metrics[(int)*t].h;
	t++;
    }
    y*=sc;
    return y;
}

void init_vbf(void)
{
    vbf_regular=load_vbf(font_regular);
    vbf_bold=load_vbf(font_bold);
}
