/*
 * (c) 2007 Stanislaw K Skowronek
 * Licensed under the terms of GPLv2 or later
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>
#include "glvbf.h"

#define MAXDEV 100
#define MAXDP 2000
#define MAXNET 8000
#define MAXNP 40

#define LINE 1024

#define XRES 2000
#define YRES 1500

#define PITCH 0.0075f

struct {
    int suspect;
    int dev[MAXNP];
    int pin[MAXNP];
    int nnp;
} nets[MAXNET];
int nnet;

int sel_mode = 0;

struct {
    int selected;
    float sel_x, sel_y, sel_a;

    float x, y, size;
    float angle, angle_c, angle_s;

    char *name;

    struct {
	float dx, dy;
	char *name;
	int net;
    } pins[MAXDP];
    int ndp;
} devices[MAXDEV];
int ndev;

void net_add_pin(int net, int dev, int pin)
{
    int i, j, jnet;
    if(net == -1)
	net=nnet++;
    if(devices[dev].pins[pin].net == net)
	return;
    if(devices[dev].pins[pin].net == -1) {
	if(nets[net].nnp>=MAXNP) {
	    fprintf(stderr, "Error: pin %d[%s].%s exceeds maximum net size\n", dev, devices[dev].name, devices[dev].pins[pin].name);
	    exit(1);
	}
	devices[dev].pins[pin].net = net;
	nets[net].dev[nets[net].nnp] = dev;
	nets[net].pin[nets[net].nnp] = pin;
	nets[net].nnp++;
	return;
    }
    jnet = devices[dev].pins[pin].net;
    if(nets[net].nnp+nets[jnet].nnp>MAXNP) {
	fprintf(stderr, "Error: pin %d[%s].%s exceeds maximum net size\n", dev, devices[dev].name, devices[dev].pins[pin].name);
	exit(1);
    }
    for(i=0,j=nets[net].nnp;i<nets[jnet].nnp;i++,j++) {
	nets[net].dev[j] = nets[jnet].dev[i];
	nets[net].pin[j] = nets[jnet].pin[i];
	devices[nets[jnet].dev[i]].pins[nets[jnet].pin[i]].net = net;
    }
    nets[net].nnp += nets[jnet].nnp;
    nets[jnet].nnp = 0;
}

char *bga_letters = "ABCDEFGHJKLMNPRTUVWY";
void parse_phys(int dev, int pin, char *name)
{
    char *p, *q;
    int px, py;
    // Xilinx "physical" pin names
    if(!strncmp(name, "IO_", 3)) {
	py = 0;
	if(!name[3])
	    return;
	for(p=name+3;*p>='A' && *p<='Z';p++) {
	    py *= 20;
	    q = strchr(bga_letters, *p);
	    if(!q)
		return;
	    py += (q-bga_letters)+1;
	}
	py--;
	px = strtol(p, &q, 10)-1;
	if(!*p || *q)
	    return;
	devices[dev].pins[pin].dx = PITCH*px;
	devices[dev].pins[pin].dy = -PITCH*py;
    }
}

void parse_pin(char *name, int *dev, int *pin)
{
    char *p, *q;
    int d, i;
    p=strchr(name, '[');
    *(p++)=0;
    q=strchr(p, ']');
    *(q++)=0;
    q++;
    d=*dev=strtol(name, NULL, 0);
    if(d>=ndev)
	ndev=d+1;
    if(!devices[d].name)
	devices[d].name=strdup(p);
    for(i=0;i<devices[d].ndp;i++)
	if(!strcmp(devices[d].pins[i].name, q)) {
	    *pin=i;
	    return;
	}
    devices[d].pins[i].name=strdup(q);
    devices[d].pins[i].net=-1;
    parse_phys(d, i, q);
    devices[d].ndp++;
    *pin=i;
}

void size_device(int dev)
{
    float x0=1.0f, y0=1.0f, x1=-1.0f, y1=-1.0f, cx, cy;
    int pin;
    if(!devices[dev].ndp)
	return;
    for(pin=0;pin<devices[dev].ndp;pin++) {
	if(devices[dev].pins[pin].dx < x0)
	    x0 = devices[dev].pins[pin].dx;
	if(devices[dev].pins[pin].dy < y0)
	    y0 = devices[dev].pins[pin].dy;
	if(devices[dev].pins[pin].dx > x1)
	    x1 = devices[dev].pins[pin].dx;
	if(devices[dev].pins[pin].dy > y1)
	    y1 = devices[dev].pins[pin].dy;
    }
    cx = (x1-x0)+PITCH*2;
    cy = (y1-y0)+PITCH*2;
    devices[dev].size = cx>cy ? cx : cy;
    cx = (x1+x0)*0.5f;
    cy = (y1+y0)*0.5f;
    for(pin=0;pin<devices[dev].ndp;pin++) {
	    devices[dev].pins[pin].dx -= cx;
	    devices[dev].pins[pin].dy -= cy;
    }
    cx = (sqrtf(devices[dev].ndp)+1.0f)*PITCH*0.8f;
    if(cx > devices[dev].size)
	devices[dev].size = cx;
    devices[dev].angle_c = 1.0f;
}

void loadnetlist(char *fname)
{
    FILE *f=fopen(fname, "r");
    char line[LINE], *p, *q;
    int d1, d2, p1, p2;
    while(fgets(line, LINE, f))
	if(line[0]>='0' && line[0]<='9') {
	    p=strchr(line, '\n');
	    if(p)
		*p=0;
	    p=strchr(line, ' ');
	    *(p++)=0;
	    q=strchr(p, ' ');
	    *(q++)=0;
	    parse_pin(line, &d1, &p1);
	    parse_pin(q, &d2, &p2);
	    if(devices[d1].pins[p1].net == -1) {
		if(devices[d2].pins[p2].net == -1) {
		    net_add_pin(-1, d1, p1);
		    net_add_pin(devices[d1].pins[p1].net, d2, p2);
		} else {
		    nets[devices[d2].pins[p2].net].suspect = 1;
		    net_add_pin(devices[d2].pins[p2].net, d1, p1);
		}
	    } else {
		if(devices[d2].pins[p2].net != -1 && devices[d2].pins[p2].net != devices[d1].pins[p1].net)
		    nets[devices[d1].pins[p1].net].suspect = 1;
		net_add_pin(devices[d1].pins[p1].net, d2, p2);
	    }
	}
    fclose(f);
    for(p1=0;p1<nnet;p1++)
	if(nets[p1].nnp && nets[p1].suspect)
	    fprintf(stderr, "Net containing pin %d[%s].%s is inconsistent\n", nets[p1].dev[0], devices[nets[p1].dev[0]].name, devices[nets[p1].dev[0]].pins[nets[p1].pin[0]].name);
    for(d1=0;d1<ndev;d1++)
	size_device(d1);
}

void vertex(int dev, unsigned int pin)
{
    float x=devices[dev].x, y=devices[dev].y;
    float dx, dy, rx, ry;
    if(pin&0x80000000) {
	rx = devices[dev].size*0.5f;
	ry = devices[dev].size*0.5f;
	if(pin&0x10)
	    rx-=PITCH*2;
	if(pin&0x20)
	    ry-=PITCH*2;
	if(pin&1)
	    dx=-rx;
	else
	    dx=rx;
	if(pin&2)
	    dy=-ry;
	else
	    dy=ry;
    } else {
	dx=devices[dev].pins[pin].dx;
	dy=devices[dev].pins[pin].dy;
    }
    rx=devices[dev].angle_c*dx - devices[dev].angle_s*dy;
    ry=devices[dev].angle_s*dx + devices[dev].angle_c*dy;
    x+=rx;
    y+=ry;
    glVertex3f(x, y, 0.0f);
}

int net_contains_sel_dev(int net)
{
    int i;
    for(i=0;i<nets[net].nnp;i++)
	if(devices[nets[net].dev[i]].selected)
	    return 1;
    return 0;
}

float get_pin_x(int dev, int pin)
{
    float dx, dy, rx;
    dx=devices[dev].pins[pin].dx;
    dy=devices[dev].pins[pin].dy;
    rx=devices[dev].angle_c*dx - devices[dev].angle_s*dy;
    return devices[dev].x+rx;
}
float get_pin_y(int dev, int pin)
{
    float dx, dy, ry;
    dx=devices[dev].pins[pin].dx;
    dy=devices[dev].pins[pin].dy;
    ry=devices[dev].angle_s*dx + devices[dev].angle_c*dy;
    return devices[dev].y+ry;
}

void draw(void)
{
    int net, i;
    char str[LINE];
    float cx, cy;
    glClear(GL_COLOR_BUFFER_BIT);
    glBegin(GL_LINES);
    for(net=0;net<nnet;net++) {
	if(net_contains_sel_dev(net))
	    glColor4ub(255, 128, 128, 192);
	else
	    glColor4ub(128, 128, 128, 64);
	if(nets[net].nnp==2) {
	    vertex(nets[net].dev[0], nets[net].pin[0]);
	    vertex(nets[net].dev[1], nets[net].pin[1]);
	} else if(nets[net].nnp>2) {
	    cx=cy=0.0f;
	    for(i=0;i<nets[net].nnp;i++) {
		cx+=get_pin_x(nets[net].dev[i], nets[net].pin[i]);
		cy+=get_pin_y(nets[net].dev[i], nets[net].pin[i]);
	    }
	    cx/=nets[net].nnp;
	    cy/=nets[net].nnp;
	    for(i=0;i<nets[net].nnp;i++) {
		glVertex3f(cx, cy, 0.0f);
		vertex(nets[net].dev[i], nets[net].pin[i]);
	    }
	}
    }
    glEnd();
    for(i=0;i<ndev;i++)
	if(devices[i].ndp) {
	    glBegin(GL_LINES);
	    if(devices[i].selected)
		glColor3ub(255, 0, 0);
	    else
		glColor3ub(255, 255, 255);
	    vertex(i, 0x80000000);
	    vertex(i, 0x80000001);
	    vertex(i, 0x80000001);
	    vertex(i, 0x80000003);
	    vertex(i, 0x80000003);
	    vertex(i, 0x80000002);
	    vertex(i, 0x80000002);
	    vertex(i, 0x80000000);
	    glEnd();
	    glBegin(GL_QUADS);
	    if(devices[i].selected)
		glColor4ub(255, 128, 128, 64);
	    else
		glColor4ub(255, 255, 255, 64);
	    vertex(i, 0x80000000);
	    vertex(i, 0x80000001);
	    vertex(i, 0x80000003);
	    vertex(i, 0x80000002);
	    vertex(i, 0x80000001);
	    vertex(i, 0x80000011);
	    vertex(i, 0x80000031);
	    vertex(i, 0x80000021);
	    glEnd();
	}
/*
    if(sel_dev!=-1) {
	sprintf(str, "%d[%s]", sel_dev, devices[sel_dev].name);
	print_vbf(0.93f-vbf_width(str, vbf_regular, 0.003f), 0.7f, str, vbf_regular, 0xff8080, 0.003f);
    }
*/
    glFlush();
    SDL_GL_SwapBuffers();
}

int container(float mx, float my, int *mode)
{
    int i;
    float dx, dy, rx, ry;
    for(i=ndev-1;i>=0;i--)
	if(devices[i].ndp) {
	    dx = mx - devices[i].x;
	    dy = my - devices[i].y;
	    rx = devices[i].angle_c*dx + devices[i].angle_s*dy;
	    ry = -devices[i].angle_s*dx + devices[i].angle_c*dy;
	    dx = fabs(rx)*2.0f;
	    dy = fabs(ry)*2.0f;
	    if(dx<devices[i].size && dy<devices[i].size) {
		if(dx<devices[i].size*0.8 && dy<devices[i].size*0.8)
		    *mode = 1;
		else
		    *mode = 2;
		return i;
	    }
	}
    *mode = 0;
    return -1;
}

void initgl(int w, int h)
{
    glViewport(0, 0, w, h);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
    glShadeModel(GL_SMOOTH);
    glScalef(1.0f, w/(float)h, 1.0f);
    init_vbf();
}

void eventloop(void)
{
    SDL_Surface *screen;
    SDL_Event event;
    int done=0,mb=0,x,y,lb,dev,ldev=0;
    float mx=0.0f, my=0.0f, lx, ly, ma;
    unsigned char *keys;
    SDL_Init(SDL_INIT_VIDEO);
    screen=SDL_SetVideoMode(XRES, YRES, 32, SDL_OPENGL);
    SDL_WM_SetCaption("JTAG Reverse Engineering", NULL);
    initgl(XRES, YRES);
    while (!done) {
	while (SDL_PollEvent(&event)) {
    	    switch(event.type) {
    		case SDL_QUIT:
        	    done=1;
            break;
    	    }
	}
	lx = mx;
	ly = my;
	lb = mb;
	mb = SDL_GetMouseState(&x, &y);
	keys=SDL_GetKeyState(NULL);
	mx = (x*2-XRES)/(float)XRES;
	my = (YRES-y*2)/(float)XRES; // not a bug - see aspect scale in initgl
	if(mb && !lb) {
	    dev = container(mx, my, &sel_mode);
	    if(dev == -1) {
		if(!keys[SDLK_RSHIFT] && !keys[SDLK_LSHIFT])
		    for(dev=0;dev<ndev;dev++)
			devices[dev].selected = 0;
	    } else {
		ldev=dev;
		if(keys[SDLK_RCTRL] || keys[SDLK_LCTRL])
		    devices[dev].selected ^= 1;
		else if(keys[SDLK_RSHIFT] || keys[SDLK_LSHIFT])
		    devices[dev].selected = 1;
		else if(!devices[dev].selected) {
		    for(dev=0;dev<ndev;dev++)
			devices[dev].selected = 0;
		    dev=ldev;
		    devices[dev].selected = 1;
		}
		if(sel_mode == 1)
		    for(dev=0;dev<ndev;dev++) {
			devices[dev].sel_x = devices[dev].x - mx;
			devices[dev].sel_y = devices[dev].y - my;
		    }
		else
		    for(dev=0;dev<ndev;dev++) {
			ma = atan2f(my - devices[ldev].y, mx - devices[ldev].x);
			devices[dev].sel_a = devices[dev].angle - ma;
		    }
	    }
	} else if(mb) {
	    if(sel_mode == 1) {
		for(dev=0;dev<ndev;dev++)
		    if(devices[dev].selected) {
			devices[dev].x = devices[dev].sel_x + mx;
			devices[dev].y = devices[dev].sel_y + my;
		    }
	    } else if (sel_mode == 2)
		for(dev=0;dev<ndev;dev++)
		    if(devices[dev].selected) {
			ma = atan2f(my - devices[ldev].y, mx - devices[ldev].x);
			devices[dev].angle = devices[dev].sel_a + ma;
			devices[dev].angle_c = cos(devices[dev].angle);
			devices[dev].angle_s = sin(devices[dev].angle);
		    }
	}
	if(keys[SDLK_ESCAPE])
    	    done=1;
	if(keys[SDLK_q])
    	    done=1;
	draw();
    }
    SDL_Quit();
}

void saveview(char *fname)
{
    int dev;
    float angle;
    FILE *f=fopen(fname, "w");
    fprintf(f, "%d\n", ndev);
    for(dev=0;dev<ndev;dev++) {
	angle = devices[dev].angle;
	while(angle<0.0f)
	    angle += 2.0f*M_PI;
	while(angle>=2.0f*M_PI)
	    angle -= 2.0f*M_PI;
	fprintf(f, "%f %f %f\n", devices[dev].x, devices[dev].y, angle);
    }
    fclose(f);
}

void loadview(char *fname)
{
    int dev, ldev;
    FILE *f=fopen(fname, "r");
    if(!f)
	return;
    fscanf(f, "%i", &ldev);
    if(ldev!=ndev) {
	fprintf(stderr, "Warning: number of devices %d mismatches saved rcfile number %d\n", ndev, ldev);
	fclose(f);
	return;
    }
    for(dev=0;dev<ldev;dev++) {
	fscanf(f, "%f%f%f", &devices[dev].x, &devices[dev].y, &devices[dev].angle);
	devices[dev].angle_c = cos(devices[dev].angle);
	devices[dev].angle_s = sin(devices[dev].angle);
    }
    fclose(f);
}

int main(int argc, char *argv[])
{
    if(argc<2) {
	fprintf(stderr, "Usage: jvis <netlist dump>\n");
	return 1;
    }
    loadnetlist(argv[1]);
    loadview("jvis.rc");
    eventloop();
    saveview("jvis.rc");
    return 0;
}
